@@ -4,6 +4,7 @@ import json
 from collections import namedtuple
 import os.path
 import numpy as np
+import cv2
 from PIL import Image, ImageDraw
 from shapely.geometry import Polygon
 from shapely.validation import explain_validity
@@ -19,6 +20,10 @@ from ocrd_utils import (
     MIME_TO_EXT
 )
 from ocrd_modelfactory import page_from_file
+from ocrd_models.ocrd_page import (
+    OrderedGroupType, OrderedGroupIndexedType,
+    RegionRefType, RegionRefIndexedType,
+)
 from ocrd import Processor
 
 from .config import OCRD_TOOL
@@ -85,7 +90,11 @@ CLASSES = {
     'NoiseRegion':                          'FF0000FF',
     'SeparatorRegion':                      'FF00FFFF',
     'UnknownRegion':                        '646464FF',
-    'CustomRegion':                         '637C81FF'}
+    'CustomRegion':                         '637C81FF',
+    'ReadingOrderLevel0':                   'DC143CFF',
+    'ReadingOrderLevel1':                   '9400D3FF',
+    'ReadingOrderLevelN':                   '8B0000FF',
+}
 # pragma pylint: enable=bad-whitespace
 
 class ExtractPages(Processor):
@@ -105,8 +114,14 @@ class ExtractPages(Processor):
         and coordinates relative to the page (which depending on the workflow could
         already be cropped, deskewed, dewarped, binarized etc). Extract the image of
         the (cropped, deskewed, dewarped) page, both in binarized form (if available)
-        and non-binarized form. In addition, create a new image with masks for each
-        segment type in ``plot_segmasks``, color-coded by class according to ``colordict``.
+        and raw form. For the latter, apply ``feature_filter`` (a comma-separated list
+        of image features, cf. :py:func:`ocrd.workspace.Workspace.image_from_page`)
+        to skip specific features when retrieving derived images. If ``transparency``
+        is true, then also add an alpha channel which is fully transparent outside of
+        the mask.
+        
+        In addition, create a new (third) image with masks for each segment type in
+        ``plot_segmasks``, color-coded by class according to ``colordict``.
         
         Create two JSON files with region types and coordinates: one (page-wise) in
         our custom format and one (global) in MS-COCO.
@@ -115,7 +130,7 @@ class ExtractPages(Processor):
         these 3 kinds of images. If fewer than 3 fileGrps are specified, they will
         share the same fileGrp (and directory). In particular, write files as follows:
         * in the first (or only) output file group (directory):
-          - ID + '.png': raw image of the (preprocessed) page
+          - ID + '.png': raw image of the page (preprocessed, but with ``feature_filter``)
           - ID + '.json': region coordinates/classes (custom format)
         * in the second (or only) output file group (directory):
           - ID + '.bin.png': binarized image of the (preprocessed) page, if available
@@ -129,7 +144,7 @@ class ExtractPages(Processor):
             where each follow-up layer and segment draws over the previous state, starting
             with a blank (white) image - unless ``plot_overlay`` is true, in which case
             each layer and segment is superimposed (alpha blended) onto the previous one,
-            starting with the raw image.
+            starting with the above raw image.
         
         In addition, write a file for all pages at once:
         * in the third (or first) output file group (directory):
@@ -192,7 +207,7 @@ class ExtractPages(Processor):
             ptype = page.get_type()
             page_image, page_coords, page_image_info = self.workspace.image_from_page(
                 page, page_id,
-                feature_filter='binarized',
+                feature_filter=self.parameter['feature_filter'],
                 transparency=self.parameter['transparency'])
             if page_image_info.resolution != 1:
                 dpi = page_image_info.resolution
@@ -317,11 +332,15 @@ class ExtractPages(Processor):
                          'bbox': [xywh['x'], xywh['y'], xywh['w'], xywh['h']],
                          'iscrowd': 0})
 
-            self.workspace.save_image_file(page_image_segmask,
-                                           file_id + '.pseg',
-                                           mask_image_grp,
-                                           page_id=page_id,
-                                           mimetype=self.parameter['mimetype'])
+            if 'order' in self.parameter['plot_segmasks']:
+                plot_order(page.get_ReadingOrder(), classes, page_image_segmask,
+                           neighbors['region'], self.parameter['plot_overlay'])
+            if self.parameter['plot_segmasks']:
+                self.workspace.save_image_file(page_image_segmask,
+                                               file_id + '.pseg',
+                                               mask_image_grp,
+                                               page_id=page_id,
+                                               mimetype=self.parameter['mimetype'])
             self.workspace.add_file(
                 ID=file_id + '.json',
                 file_grp=mask_image_grp,
@@ -390,6 +409,54 @@ def segment_poly(page_id, segment, coords):
         return None
     return poly
 
+def plot_order(readingorder, classes, image, regions, alpha=False):
+    LOG = getLogger('processor.ExtractPages')
+    regiondict = dict((region.id, region.poly) for region in regions)
+    def get_points(rogroup, level):
+        points = list()
+        if isinstance(rogroup, (OrderedGroupType, OrderedGroupIndexedType)):
+            # FIXME: @index is broken in prima-core-libs+prima-page-viewer and producers
+            # (so we have to do ignore index here too to stay compatible)
+            regionrefs = rogroup.get_AllIndexed(index_sort=False)
+        else:
+            # FIXME: PageViewer does not render these in-order via arrows,
+            #        but creates a "star" plus circle for unordered groups
+            regionrefs = rogroup.get_UnorderedGroupChildren()
+        for regionref in regionrefs:
+            morepoints = list()
+            poly = regiondict.get(regionref.get_regionRef(), None)
+            if poly:
+                # we have seen this region
+                morepoints.append((level, tuple(np.array(poly.centroid, np.int))))
+            if not isinstance(regionref, (RegionRefType, RegionRefIndexedType)):
+                # try to get subgroup regions instead
+                morepoints = get_points(regionref, level + 1) or morepoints
+            points.extend(morepoints)
+        return points
+    newimg = 255 * np.ones((image.height, image.width, 3), np.uint8)
+    points = [(0, (0, 0))]
+    if readingorder:
+        readingorder = readingorder.get_OrderedGroup() or readingorder.get_UnorderedGroup()
+    if readingorder:
+        # use recursive group ordering
+        points.extend(get_points(readingorder, 0))
+    else:
+        # use XML ordering
+        points.extend([(0, tuple(np.array(region.poly.centroid, np.int))) for region in regions])
+    for p1, p2 in zip(points[:-1], points[1:]):
+        color = 'ReadingOrderLevel%s' % (str(p1[0]) if p1[0] < 2 else 'N')
+        if color not in classes:
+            LOG.error('mask plots requested, but "colordict" does not contain a "%s" mapping', color)
+            return
+        color = classes[color]
+        color = (int(color[0:2], 16),
+                 int(color[2:4], 16),
+                 int(color[4:6], 16))
+        cv2.arrowedLine(newimg, p1[1], p2[1], color, thickness=2, tipLength=0.01)
+    layer = Image.fromarray(newimg)
+    layer.putalpha(Image.fromarray(255 * np.any(newimg < 255, axis=2).astype(np.uint8), mode='L'))
+    image.alpha_composite(layer)
+
 def plot_segment(page_id, segment, poly, stype, classes, image, neighbors, alpha=False):
     LOG = getLogger('processor.ExtractPages')
     if not poly:
@@ -417,6 +484,7 @@ def plot_segment(page_id, segment, poly, stype, classes, image, neighbors, alpha
                 LOG.warning('Page "%s" segment "%s" within neighbour "%s" (IoU: %.3f)',
                             page_id, segment.id, neighbor.id,
                             poly.area / neighbor.poly.area)
+    if segment and hasattr(segment, 'id'):
         neighbors.append(Neighbor(segment.id, poly, stype))
     # draw segment
     if alpha:
